//
//  Ribbon.swift
//  gloss
//
//  Created by Saint on 2/24/23.
//

import GRDB
/// The Line struct.
///
/// Identifiable conformance supports SwiftUI list animations, and type-safe
/// GRDB primary key methods.
/// Equatable conformance supports tests.
struct Ribbon: Identifiable, Equatable {
    /// The player id.
    ///
    /// Int64 is the recommended type for auto-incremented database ids.
    /// Use nil for players that are not inserted yet in the database.
    var id: Int64?
    var groupId: Int
    var pos: Int
    var undoLevel: Int
    var currentLevel: Int
    var minLevel: Int
    var maxLevel: Int
    var title: String
    var book: String
    var scrollId: String
    var scrollOffset: Int
}

extension Ribbon {
}


/// Make Line a Codable Record.
///
/// See <https://github.com/groue/GRDB.swift/blob/master/README.md#records>
extension Ribbon: Codable, FetchableRecord, MutablePersistableRecord {
    // Define database columns from CodingKeys
    fileprivate enum Columns {
        static let id = Column(CodingKeys.id)
        static let pos = Column(CodingKeys.pos)
        static let book = Column(CodingKeys.book)
        static let title = Column(CodingKeys.title)
        static let scrollOffset = Column(CodingKeys.scrollOffset)
    }

    /// Updates a player id after it has been inserted in the database.
    mutating func didInsert(_ inserted: InsertionSuccess) {
        id = inserted.rowID
    }
}

// MARK: - Line Database Requests

/// Define some player requests used by the application.
///
/// See <https://github.com/groue/GRDB.swift/blob/master/README.md#requests>
/// See <https://github.com/groue/GRDB.swift/blob/master/Documentation/GoodPracticesForDesigningRecordTypes.md>
extension DerivableRequest<Ribbon> {
}
